/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef _GMRF_MMSS_MMSS_HPP_
#define _GMRF_MMSS_MMSS_HPP_

#include "queue.hpp"

#include <memory>


namespace MMSS {

class context_t {
public:
	uint64_t now;
	timeout_queue_t<packet_t> packet_queue;
	timeout_queue_t<scheduled_t> scheduled_queue;

	context_t() : now(0), packet_queue(this), scheduled_queue(this) {}
};

class config_t {
public:
	network_t *networks;
	gmrf_t *nodes;
};

class network_t {
public:
	network_t *next;

	char *name;

	gmrf_iface_t *interfaces;
	size_t mtu;
};

class packet_t {
public:
	uint64_t sent;

	gmrf_iface_t *source;
	gmrf_iface_t *dest;

	size_t len;
	std::unique_ptr<uint8_t> data;
};

class scheduled_t {
public:
	gmrf_t *node;
	gmrf_scheduled_func f;
	void *arg;
};


const protocol_t* load_protocol(const char *module);

namespace Config {

void add_network(context_t *mmss, config_t *conf, const char *name);

}

bool read_config(context_t *mmss, config_t *conf, const char *filename);

void add_iface(gmrf_t *node, network_t *net, const char *name, const gmrf_addr_t *address);

uint64_t now(const context_t *mmss);

void dispatch(std::shared_ptr<packet_t> packet);
void run_scheduled(std::shared_ptr<scheduled_t> scheduled);

void logf(context_t *mmss, int priority, const char *format, ...);


static inline int max(int a, int b) {
	return (a > b) ? a : b;
}

static inline int min(int a, int b) {
	return (a < b) ? a : b;
}


static inline size_t alignto(size_t l, size_t a) {
	return ((l+a-1)/a)*a;
}

}

struct gmrf {
	gmrf_t *next;

	char *name;

	MMSS::context_t *mmss;
	gmrf_context_t *ctx;
	gmrf_iface_t *interfaces;

	unsigned rand_seed;

	const MMSS::protocol_t *proto;
};

struct gmrf_iface {
	gmrf_iface_t *node_next;
	gmrf_iface_t *network_next;

	char *name;
	gmrf_addr_t address;

	gmrf_t *node;
	MMSS::network_t *net;
};

#endif /* _GMRF_MMSS_MMSS_HPP_ */
